// Code generated by stringer -i a.out.go -o anames.go -p loong64; DO NOT EDIT.

package loong64

import "cmd/internal/obj"

var Anames = []string{
	obj.A_ARCHSPECIFIC: "ABSD",
	"ABSF",
	"ADD",
	"ADDD",
	"ADDF",
	"ADDU",
	"ADDW",
	"AND",
	"BEQ",
	"BGEZ",
	"BLEZ",
	"BGTZ",
	"BLTZ",
	"BFPF",
	"BFPT",
	"BNE",
	"BREAK",
	"CMPEQD",
	"CMPEQF",
	"CMPGED",
	"CMPGEF",
	"CMPGTD",
	"CMPGTF",
	"LU12IW",
	"LU32ID",
	"LU52ID",
	"PCALAU12I",
	"PCADDU12I",
	"JIRL",
	"BGE",
	"BLT",
	"BLTU",
	"BGEU",
	"DIV",
	"DIVD",
	"DIVF",
	"DIVU",
	"DIVW",
	"LL",
	"LLV",
	"LUI",
	"MOVB",
	"MOVBU",
	"MOVD",
	"MOVDF",
	"MOVDW",
	"MOVF",
	"MOVFD",
	"MOVFW",
	"MOVH",
	"MOVHU",
	"MOVW",
	"MOVWD",
	"MOVWF",
	"MUL",
	"MULD",
	"MULF",
	"MULU",
	"MULH",
	"MULHU",
	"MULW",
	"NEGD",
	"NEGF",
	"NEGW",
	"NEGV",
	"NOOP",
	"NOR",
	"OR",
	"REM",
	"REMU",
	"RFE",
	"SC",
	"SCV",
	"SGT",
	"SGTU",
	"SLL",
	"SQRTD",
	"SQRTF",
	"SRA",
	"SRL",
	"ROTR",
	"SUB",
	"SUBD",
	"SUBF",
	"SUBU",
	"SUBW",
	"DBAR",
	"SYSCALL",
	"TEQ",
	"TNE",
	"WORD",
	"XOR",
	"MASKEQZ",
	"MASKNEZ",
	"MOVV",
	"SLLV",
	"SRAV",
	"SRLV",
	"ROTRV",
	"DIVV",
	"DIVVU",
	"REMV",
	"REMVU",
	"MULV",
	"MULVU",
	"MULHV",
	"MULHVU",
	"ADDV",
	"ADDVU",
	"SUBV",
	"SUBVU",
	"TRUNCFV",
	"TRUNCDV",
	"TRUNCFW",
	"TRUNCDW",
	"MOVWU",
	"MOVFV",
	"MOVDV",
	"MOVVF",
	"MOVVD",
	"ORN",
	"ANDN",
	"AMSWAPB",
	"AMSWAPH",
	"AMSWAPW",
	"AMSWAPV",
	"AMCASB",
	"AMCASH",
	"AMCASW",
	"AMCASV",
	"AMADDW",
	"AMADDV",
	"AMANDW",
	"AMANDV",
	"AMORW",
	"AMORV",
	"AMXORW",
	"AMXORV",
	"AMMAXW",
	"AMMAXV",
	"AMMINW",
	"AMMINV",
	"AMMAXWU",
	"AMMAXVU",
	"AMMINWU",
	"AMMINVU",
	"AMSWAPDBB",
	"AMSWAPDBH",
	"AMSWAPDBW",
	"AMSWAPDBV",
	"AMCASDBB",
	"AMCASDBH",
	"AMCASDBW",
	"AMCASDBV",
	"AMADDDBW",
	"AMADDDBV",
	"AMANDDBW",
	"AMANDDBV",
	"AMORDBW",
	"AMORDBV",
	"AMXORDBW",
	"AMXORDBV",
	"AMMAXDBW",
	"AMMAXDBV",
	"AMMINDBW",
	"AMMINDBV",
	"AMMAXDBWU",
	"AMMAXDBVU",
	"AMMINDBWU",
	"AMMINDBVU",
	"EXTWB",
	"EXTWH",
	"CLOW",
	"CLOV",
	"CLZW",
	"CLZV",
	"CTOW",
	"CTOV",
	"CTZW",
	"CTZV",
	"REVBV",
	"REVB2W",
	"REVB4H",
	"REVB2H",
	"REVH2W",
	"REVHV",
	"BITREV4B",
	"BITREV8B",
	"BITREVW",
	"BITREVV",
	"BSTRINSW",
	"BSTRINSV",
	"BSTRPICKW",
	"BSTRPICKV",
	"CRCWBW",
	"CRCWHW",
	"CRCWWW",
	"CRCWVW",
	"CRCCWBW",
	"CRCCWHW",
	"CRCCWWW",
	"CRCCWVW",
	"RDTIMELW",
	"RDTIMEHW",
	"RDTIMED",
	"CPUCFG",
	"FMADDF",
	"FMADDD",
	"FMSUBF",
	"FMSUBD",
	"FNMADDF",
	"FNMADDD",
	"FNMSUBF",
	"FNMSUBD",
	"FMINF",
	"FMIND",
	"FMAXF",
	"FMAXD",
	"FCOPYSGF",
	"FCOPYSGD",
	"FSCALEBF",
	"FSCALEBD",
	"FLOGBF",
	"FLOGBD",
	"FCLASSF",
	"FCLASSD",
	"FFINTFW",
	"FFINTFV",
	"FFINTDW",
	"FFINTDV",
	"FTINTWF",
	"FTINTWD",
	"FTINTVF",
	"FTINTVD",
	"FTINTRPWF",
	"FTINTRPWD",
	"FTINTRPVF",
	"FTINTRPVD",
	"FTINTRMWF",
	"FTINTRMWD",
	"FTINTRMVF",
	"FTINTRMVD",
	"FTINTRZWF",
	"FTINTRZWD",
	"FTINTRZVF",
	"FTINTRZVD",
	"FTINTRNEWF",
	"FTINTRNEWD",
	"FTINTRNEVF",
	"FTINTRNEVD",
	"VMOVQ",
	"XVMOVQ",
	"VPCNTB",
	"VPCNTH",
	"VPCNTW",
	"VPCNTV",
	"XVPCNTB",
	"XVPCNTH",
	"XVPCNTW",
	"XVPCNTV",
	"VSEQB",
	"XVSEQB",
	"VSEQH",
	"XVSEQH",
	"VSEQW",
	"XVSEQW",
	"VSEQV",
	"XVSEQV",
	"LAST",
}
